/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Lustre Light Super operations
 *
 * This code is issued under the GNU General Public License.
 * See the file COPYING in this distribution
 *
 * Copryright (C) 2002 Cluster File Systems, Inc.
 */

#define DEBUG_SUBSYSTEM S_LLITE

#include <linux/module.h>
#include <linux/lustre_lite.h>
#include <linux/lustre_ha.h>
#include <linux/lustre_dlm.h>

kmem_cache_t *ll_file_data_slab;
extern struct address_space_operations ll_aops;
extern struct address_space_operations ll_dir_aops;
struct super_operations ll_super_operations;

extern int ll_recover(struct ptlrpc_client *);
extern int ll_commitcbd_setup(struct ll_sb_info *);
extern int ll_commitcbd_cleanup(struct ll_sb_info *);

static char *ll_read_opt(const char *opt, char *data)
{
        char *value;
        char *retval;
        ENTRY;

        CDEBUG(D_SUPER, "option: %s, data %s\n", opt, data);
        if ( strncmp(opt, data, strlen(opt)) )
                RETURN(NULL);
        if ( (value = strchr(data, '=')) == NULL )
                RETURN(NULL);

        value++;
        OBD_ALLOC(retval, strlen(value) + 1);
        if ( !retval ) {
                CERROR("out of memory!\n");
                RETURN(NULL);
        }
        
        memcpy(retval, value, strlen(value)+1);
        CDEBUG(D_SUPER, "Assigned option: %s, value %s\n", opt, retval);
        RETURN(retval);
}

static void ll_options(char *options, char **ost, char **mds)
{
        char *this_char;
        ENTRY;

        if (!options) {
                EXIT;
                return;
        }

        for (this_char = strtok (options, ",");
             this_char != NULL;
             this_char = strtok (NULL, ",")) {
                CDEBUG(D_SUPER, "this_char %s\n", this_char);
                if ( (!*ost && (*ost = ll_read_opt("ost", this_char)))||
                     (!*mds && (*mds = ll_read_opt("mds", this_char))) )
                        continue;
        }
        EXIT;
}

#ifndef log2
#define log2(n) ffz(~(n))
#endif

static struct super_block * ll_read_super(struct super_block *sb,
                                          void *data, int silent)
{
        struct inode *root = 0;
        struct obd_device *obd;
        struct ll_sb_info *sbi;
        char *ost = NULL;
        char *mds = NULL;
        int err;
        struct ll_fid rootfid;
        struct statfs sfs;
        __u64 last_committed, last_rcvd;
        __u32 last_xid;
        struct ptlrpc_request *request = NULL;
        struct ll_inode_md md;

        ENTRY;
        MOD_INC_USE_COUNT;

        OBD_ALLOC(sbi, sizeof(*sbi));
        if (!sbi) {
                MOD_DEC_USE_COUNT;
                RETURN(NULL);
        }

        sb->u.generic_sbp = sbi;

        ll_options(data, &ost, &mds);

        if (!ost) {
                CERROR("no ost\n");
                GOTO(out_free, sb = NULL);
        }

        if (!mds) {
                CERROR("no mds\n");
                GOTO(out_free, sb = NULL);
        }

        obd = class_uuid2obd(mds); 
        if (!obd) {
                CERROR("MDS %s: not setup or attached\n", mds);
                GOTO(out_free, sb = NULL);
        }

        /* First the MDS since an LOV requires an the MDC connection 
           to find its descriptor */ 
#if 0
        err = connmgr_connect(ptlrpc_connmgr, sbi->ll_mds_conn);
        if (err) {
                CERROR("cannot connect to MDS: rc = %d\n", err);
                GOTO(out_rpc, sb = NULL);
        }
#endif 

        err = obd_connect(&sbi->ll_mdc_conn, obd);
        if (err) {
                CERROR("cannot connect to %s: rc = %d\n", mds, err);
                GOTO(out_free, sb = NULL);
        }
        sbi2mdc(sbi)->mdc_conn->c_level = LUSTRE_CONN_FULL;

        /* now the OST, which could be an LOV */ 

        obd = class_uuid2obd(ost);
        if (!obd) {
                CERROR("OST %s: not setup or attached\n", ost);
                GOTO(out_free, sb = NULL);
        }
        err = obd_connect(&sbi->ll_osc_conn, obd);
        if (err) {
                CERROR("cannot connect to %s: rc = %d\n", ost, err);
                GOTO(out_free, sb = NULL);
        }

        /* XXX: need to store the last_* values somewhere */
        err = mdc_getstatus(&sbi->ll_mdc_conn,
                            &rootfid, &last_committed, 
                            &last_rcvd, &last_xid,
                            &request);
        if (err) {
                CERROR("cannot mds_connect: rc = %d\n", err);
                GOTO(out_disc, sb = NULL);
        }
        CDEBUG(D_SUPER, "rootfid %Ld\n", (unsigned long long)rootfid.id);
        sbi->ll_rootino = rootfid.id;

        memset(&sfs, 0, sizeof(sfs));
        err = mdc_statfs(&sbi->ll_mdc_conn, &sfs, &request);
        sb->s_blocksize = sfs.f_bsize;
        sb->s_blocksize_bits = log2(sfs.f_bsize);
        sb->s_magic = LL_SUPER_MAGIC;
        sb->s_maxbytes = (1ULL << (32 + 9)) - sfs.f_bsize;
        ptlrpc_req_finished(request);

        sb->s_op = &ll_super_operations;

        /* make root inode */
        err = mdc_getattr(&sbi->ll_mdc_conn,
                          sbi->ll_rootino, S_IFDIR, OBD_MD_FLNOTOBD|OBD_MD_FLBLOCKS, 0, &request);
        if (err) {
                CERROR("mdc_getattr failed for root: rc = %d\n", err);
                GOTO(out_mdc, sb = NULL);
        }

        /* initialize committed transaction callback daemon */
        spin_lock_init(&sbi->ll_commitcbd_lock); 
        init_waitqueue_head(&sbi->ll_commitcbd_waitq);
        init_waitqueue_head(&sbi->ll_commitcbd_ctl_waitq);
        sbi->ll_commitcbd_flags = 0;
        err = ll_commitcbd_setup(sbi);
        if (err) {
                CERROR("failed to start commit callback daemon: rc = %d\n",err);
                GOTO(out_mdc, sb = NULL);
        }

        md.body = lustre_msg_buf(request->rq_repmsg, 0);
        md.md = NULL;
        root = iget4(sb, sbi->ll_rootino, NULL, &md);
                     
        if (root) {
                sb->s_root = d_alloc_root(root);
        } else {
                CERROR("lustre_lite: bad iget4 for root\n");
                GOTO(out_cdb, sb = NULL);
        }

        ptlrpc_free_req(request);

out_dev:
        if (mds)
                OBD_FREE(mds, strlen(mds) + 1);
        if (ost)
                OBD_FREE(ost, strlen(ost) + 1);

        RETURN(sb);

out_cdb:
        ll_commitcbd_cleanup(sbi);
out_mdc:
        obd_disconnect(&sbi->ll_mdc_conn);
        obd_disconnect(&sbi->ll_osc_conn);
out_disc:
        ptlrpc_free_req(request);
        obd_disconnect(&sbi->ll_osc_conn);
out_free:
        OBD_FREE(sbi, sizeof(*sbi));

        MOD_DEC_USE_COUNT;
        goto out_dev;
} /* ll_read_super */

static void ll_put_super(struct super_block *sb)
{
        struct ll_sb_info *sbi = ll_s2sbi(sb);
        ENTRY;
        ll_commitcbd_cleanup(sbi);
        obd_disconnect(&sbi->ll_osc_conn);
        obd_disconnect(&sbi->ll_mdc_conn);
        OBD_FREE(sbi, sizeof(*sbi));

        MOD_DEC_USE_COUNT;
        EXIT;
} /* ll_put_super */

static void ll_clear_inode(struct inode *inode)
{
        if (atomic_read(&inode->i_count) == 0) {
                struct lov_stripe_md *md = ll_i2info(inode)->lli_smd;
                if (md) {
                        OBD_FREE(md, md->lmd_size); 
                        ll_i2info(inode)->lli_smd = NULL;
                }
                if (ll_i2info(inode)->lli_symlink_name) {
                        OBD_FREE(ll_i2info(inode)->lli_symlink_name,
                                 strlen(ll_i2info(inode)->lli_symlink_name)+ 1);
                        ll_i2info(inode)->lli_symlink_name = NULL;
                }
        }
}

static void ll_delete_inode(struct inode *inode)
{
        if (S_ISREG(inode->i_mode)) { 
                int err;
                struct obdo oa;
                struct lov_stripe_md *md = ll_i2info(inode)->lli_smd;
 
               if (!md)
                        GOTO(out, -EINVAL);

                oa.o_id = md->lmd_object_id;
                oa.o_easize = md->lmd_size;
                if (oa.o_id == 0) { 
                        CERROR("This really happens\n"); 
                        /* No obdo was ever created */
                        GOTO(out, 0);
                }

                err = obd_destroy(ll_i2obdconn(inode), &oa, md);
                CDEBUG(D_SUPER, "obd destroy of %Ld error %d\n",
                       md->lmd_object_id, err);
        }
out:
        clear_inode(inode);
}

/* like inode_setattr, but doesn't mark the inode dirty */
static int ll_attr2inode(struct inode * inode, struct iattr * attr, int trunc)
{
        unsigned int ia_valid = attr->ia_valid;
        int error = 0;

        if ((ia_valid & ATTR_SIZE) && trunc) {
                error = vmtruncate(inode, attr->ia_size);
                if (error)
                        goto out;
        } else if (ia_valid & ATTR_SIZE)
                inode->i_size = attr->ia_size;

        if (ia_valid & ATTR_UID)
                inode->i_uid = attr->ia_uid;
        if (ia_valid & ATTR_GID)
                inode->i_gid = attr->ia_gid;
        if (ia_valid & ATTR_ATIME)
                inode->i_atime = attr->ia_atime;
        if (ia_valid & ATTR_MTIME)
                inode->i_mtime = attr->ia_mtime;
        if (ia_valid & ATTR_CTIME)
                inode->i_ctime = attr->ia_ctime;
        if (ia_valid & ATTR_MODE) {
                inode->i_mode = attr->ia_mode;
                if (!in_group_p(inode->i_gid) && !capable(CAP_FSETID))
                        inode->i_mode &= ~S_ISGID;
        }
out:
        return error;
}

int ll_inode_setattr(struct inode *inode, struct iattr *attr, int do_trunc)
{
        struct ptlrpc_request *request = NULL;
        struct ll_sb_info *sbi = ll_i2sbi(inode);
        int err;

        ENTRY;

        /* change incore inode */
        ll_attr2inode(inode, attr, do_trunc);

        err = mdc_setattr(&sbi->ll_mdc_conn, inode, attr, &request);
        if (err)
                CERROR("mdc_setattr fails (%d)\n", err);

        ptlrpc_req_finished(request);

        RETURN(err);
}

int ll_setattr(struct dentry *de, struct iattr *attr)
{
        int rc = inode_change_ok(de->d_inode, attr);

        if (rc)
                return rc;

        return ll_inode_setattr(de->d_inode, attr, 1);
}

static int ll_statfs(struct super_block *sb, struct statfs *sfs)
{
        struct ptlrpc_request *request = NULL;
        struct ll_sb_info *sbi = ll_s2sbi(sb);
        int rc;
        ENTRY;

        memset(sfs, 0, sizeof(*sfs));
        rc = mdc_statfs(&sbi->ll_mdc_conn, sfs, &request);
        ptlrpc_req_finished(request);
        if (rc)
                CERROR("obd_statfs fails: rc = %d\n", rc);
        else
                CDEBUG(D_SUPER, "statfs shows blocks %ld/%ld objects %ld/%ld\n",
                       sfs->f_bavail, sfs->f_blocks, sfs->f_files,sfs->f_ffree);

        /* temporary until mds_statfs returns statfs info for all OSTs */
        if (!rc) {
                struct statfs obd_sfs;

                rc = obd_statfs(&sbi->ll_osc_conn, &obd_sfs);
                if (rc) {
                        CERROR("obd_statfs fails: rc = %d\n", rc);
                        GOTO(out, rc);
                }
                CDEBUG(D_SUPER, "obd_statfs returns blocks %ld/%ld, "
                       "objects %ld/%ld\n",
                       obd_sfs.f_bavail, obd_sfs.f_blocks,
                       obd_sfs.f_ffree, obd_sfs.f_files);

                sfs->f_bfree = obd_sfs.f_bfree;
                sfs->f_bavail = obd_sfs.f_bavail;
                sfs->f_blocks = obd_sfs.f_blocks;
                if (obd_sfs.f_ffree < sfs->f_ffree)
                        sfs->f_ffree = obd_sfs.f_ffree;
        }

out:
        RETURN(rc);
}

inline int ll_stripe_md_size(struct super_block *sb)
{
        struct mdc_obd *mdc = sbi2mdc(ll_s2sbi(sb));
        return mdc->mdc_max_mdsize;
}

static void ll_to_inode(struct inode *dst, struct ll_inode_md *md)
{
        struct mds_body *body = md->body;
        struct ll_inode_info *ii = ll_i2info(dst);

        /* core attributes first */
        if (body->valid & OBD_MD_FLID)
                dst->i_ino = body->ino;
        if (body->valid & OBD_MD_FLATIME)
                dst->i_atime = body->atime;
        if (body->valid & OBD_MD_FLMTIME)
                dst->i_mtime = body->mtime;
        if (body->valid & OBD_MD_FLCTIME)
                dst->i_ctime = body->ctime;
        if (body->valid & OBD_MD_FLSIZE)
                dst->i_size = body->size;
        if (body->valid & OBD_MD_FLMODE)
                dst->i_mode = body->mode;
        if (body->valid & OBD_MD_FLUID)
                dst->i_uid = body->uid;
        if (body->valid & OBD_MD_FLGID)
                dst->i_gid = body->gid;
        if (body->valid & OBD_MD_FLFLAGS)
                dst->i_flags = body->flags;
        if (body->valid & OBD_MD_FLNLINK)
                dst->i_nlink = body->nlink;
        if (body->valid & OBD_MD_FLGENER)
                dst->i_generation = body->generation;
        if (body->valid & OBD_MD_FLRDEV)
                dst->i_rdev = body->extra;
        if (md && md->md && md->md->lmd_stripe_count) { 
                struct lov_stripe_md *smd = md->md;
                int size = ll_stripe_md_size(dst->i_sb);
                if (md->md->lmd_size != size) { 
                        CERROR("Striping metadata size error %ld\n",
                               dst->i_ino); 
                        LBUG();
                }
                OBD_ALLOC(ii->lli_smd, size);
                if (!ii->lli_smd){ 
                        CERROR("No memory for %d\n", size);
                        LBUG();
                }
                memcpy(ii->lli_smd, smd, size);
        }
} /* ll_to_inode */

static void ll_read_inode2(struct inode *inode, void *opaque)
{
        struct ll_inode_md *md = opaque;

        ENTRY;
        ll_to_inode(inode, md);

        /* OIDEBUG(inode); */

        if (S_ISREG(inode->i_mode)) {
                inode->i_op = &ll_file_inode_operations;
                inode->i_fop = &ll_file_operations;
                inode->i_mapping->a_ops = &ll_aops;
                EXIT;
        } else if (S_ISDIR(inode->i_mode)) {
                inode->i_op = &ll_dir_inode_operations;
                inode->i_fop = &ll_dir_operations;
                inode->i_mapping->a_ops = &ll_dir_aops;
                EXIT;
        } else if (S_ISLNK(inode->i_mode)) {
                inode->i_op = &ll_fast_symlink_inode_operations;
                EXIT;
        } else {
                init_special_inode(inode, inode->i_mode,
                                   ((int *)ll_i2info(inode)->lli_inline)[0]);
                EXIT;
        }

        return;
}

/* exported operations */
struct super_operations ll_super_operations =
{
        read_inode2: ll_read_inode2,
        clear_inode: ll_clear_inode,
        delete_inode: ll_delete_inode,
        put_super: ll_put_super,
        statfs: ll_statfs
};

struct file_system_type lustre_lite_fs_type = {
        "lustre_lite", 0, ll_read_super, NULL
};

static int __init init_lustre_lite(void)
{
        printk(KERN_INFO "Lustre Lite 0.0.1, info@clusterfs.com\n");
        ll_file_data_slab = kmem_cache_create("ll_file_data",
                                              sizeof(struct ll_file_data), 0,
                                               SLAB_HWCACHE_ALIGN, NULL, NULL);
        if (ll_file_data_slab == NULL)
                return -ENOMEM;
        return register_filesystem(&lustre_lite_fs_type);
}

static void __exit exit_lustre_lite(void)
{
        unregister_filesystem(&lustre_lite_fs_type);
        kmem_cache_destroy(ll_file_data_slab);
}

MODULE_AUTHOR("Cluster File Systems, Inc. <info@clusterfs.com>");
MODULE_DESCRIPTION("Lustre Lite Client File System v1.0");
MODULE_LICENSE("GPL");

module_init(init_lustre_lite);
module_exit(exit_lustre_lite);
